module AMR exposing (application, Program)

{-| This module helps you set up an Elm `Program` for development.

Automatic message-replay solves the same problem as "hot-reloading" - in the sense that your application won't lose state during development. It does however differ in that, since the interactions are recorded rather than the state itself, you get to change virtually any part of your program without breaking replay.

Interactions recorded is based on the `message`-value that [the elm-architecture](https://guide.elm-lang.org/architecture/) employs. If you remove or modify any of these, you will obviously change the behavior of message-replay - This module will replay all message up until the first that is unrecognized.

@docs application, Program

-}

import Browser
import Browser.Navigation
import DenseList exposing (DenseList)
import Html
import Json.Decode
import Json.Encode
import Url exposing (Url)



-- API


{-| A program with automatic message-replay enabled.

A nice trick to toggle between this kind of program and the production-build
is to insert both constructors side-by-side and commenting out the one your currently using

-}
type alias Program flags model msg =
    Platform.Program flags (Model model) (Msg msg)


{-| This API mirrors `elm/browser`.

You can find documentation for [here](https://package.elm-lang.org/packages/elm/browser/latest/) if you don't know it yet!

-}
application :
    { encodeMsg : msg -> Json.Encode.Value
    , msgDecoder : Json.Decode.Decoder msg
    , fromCache : flags -> Maybe String
    , toCache : String -> Cmd (Msg msg)
    }
    ->
        { init : flags -> Url -> Browser.Navigation.Key -> ( model, Cmd msg )
        , update : msg -> model -> ( model, Cmd msg )
        , subscriptions : model -> Sub msg
        , view : model -> Browser.Document msg
        , onUrlChange : Url -> msg
        , onUrlRequest : Browser.UrlRequest -> msg
        }
    -> Program flags model msg
application config app =
    Browser.application
        { update = update app.update config.toCache config.encodeMsg
        , subscriptions = Sub.map UpdateApp << app.subscriptions << .model
        , view = mapDocument UpdateApp << app.view << .model
        , onUrlChange = UpdateApp << app.onUrlChange
        , onUrlRequest = UpdateApp << app.onUrlRequest
        , init =
            \flags url key ->
                init app.update
                    config.msgDecoder
                    (config.fromCache flags)
                    (app.init flags url key)
        }


mapDocument : (a -> b) -> Browser.Document a -> Browser.Document b
mapDocument msg { title, body } =
    { title = title
    , body = List.map (Html.map msg) body
    }



-- Program


type alias Model model =
    { model : model
    , msgs : DenseList String
    }


type Msg msg
    = UpdateApp msg


init :
    (msg -> model -> ( model, Cmd msg ))
    -> Json.Decode.Decoder msg
    -> Maybe String
    -> ( model, Cmd msg )
    -> ( Model model, Cmd (Msg msg) )
init updateApp msgDecoder fromCache ( model, appCmd ) =
    case Maybe.map (Json.Decode.decodeString (decoder (withoutCmd updateApp) msgDecoder model)) fromCache of
        Just (Ok cacheModel) ->
            ( cacheModel
            , Cmd.none
            )

        _ ->
            ( { model = model
              , msgs = DenseList.empty
              }
            , Cmd.map UpdateApp appCmd
            )


update :
    (msg -> model -> ( model, Cmd msg ))
    -> (String -> Cmd (Msg msg))
    -> (msg -> Json.Encode.Value)
    -> Msg msg
    -> Model model
    -> ( Model model, Cmd (Msg msg) )
update updateApp toCache encodeMsg msg model =
    case msg of
        UpdateApp updateMsg ->
            let
                ( updateModel, appCmd ) =
                    updateApp updateMsg model.model

                updatedModel =
                    { model
                        | model = updateModel
                        , msgs = DenseList.insert (Json.Encode.encode 0 (encodeMsg updateMsg)) model.msgs
                    }
            in
            ( updatedModel
            , Cmd.batch
                [ Cmd.map UpdateApp appCmd
                , toCache (Json.Encode.encode 0 (encode updatedModel))
                ]
            )


encode : Model model -> Json.Encode.Value
encode model =
    DenseList.encodeReversed Json.Encode.string model.msgs


decoder :
    (msg -> model -> model)
    -> Json.Decode.Decoder msg
    -> model
    -> Json.Decode.Decoder (Model model)
decoder replayUpdate msgDecoder model =
    Json.Decode.map
        (\( msgs, updatedModel ) ->
            { model = updatedModel
            , msgs = msgs
            }
        )
        (DenseList.foldLeftUntilErrorDecoder
            Json.Decode.string
            (Json.Decode.decodeString msgDecoder)
            replayUpdate
            model
        )


withoutCmd : (msg -> model -> ( model, Cmd msg )) -> msg -> model -> model
withoutCmd updateApp msg model =
    Tuple.first (updateApp msg model)
